import { parseComponent, isVue3, isCodeVueSfc } from 'vue-inbrowser-compiler-utils';
export { adaptCreateElement, addScopedStyle, concatenate, isCodeVueSfc } from 'vue-inbrowser-compiler-utils';
import { transform } from 'sucrase';
import walkes from 'walkes';
import { parse as parse$1 } from '@babel/parser';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __spreadArray(to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
}

var babelParserOptions = {
    sourceType: 'module',
    strictMode: false,
    tokens: true,
    plugins: [
        'jsx',
        'decorators-legacy',
        'doExpressions',
        'objectRestSpread',
        'classProperties',
        'classPrivateProperties',
        'classPrivateMethods',
        'exportDefaultFrom',
        'exportNamespaceFrom',
        'asyncGenerators',
        'functionBind',
        'functionSent',
        'dynamicImport',
        'numericSeparator',
        'optionalChaining',
        'importMeta',
        'bigInt',
        'optionalCatchBinding',
        'throwExpressions',
        'nullishCoalescingOperator',
        'typescript'
    ]
};
function buildParse(options) {
    if (options === void 0) { options = {}; }
    options = __assign(__assign({}, babelParserOptions), options);
    return {
        parse: function (src) {
            return parse$1(src, options);
        }
    };
}

var parse;
function getAst(code) {
    parse = parse !== null && parse !== void 0 ? parse : buildParse().parse;
    return parse(code);
}

function transformOneJSXSpread(node, code, offset, config) {
    var start = node.start + offset;
    var end = node.end + offset;
    var objectAssign = config.objectAssign || 'Object.assign';
    var statement = code.substring(start, end);
    var transpiledStatement = "<".concat(node.name.name, " {...").concat(objectAssign, "(").concat(node.attributes
        .map(function (attrNode) {
        if (attrNode.type === 'JSXSpreadAttribute') {
            return "".concat(code.substring(attrNode.argument.start + offset, attrNode.argument.end + offset));
        }
        else {
            return "{".concat(attrNode.name.name, ":").concat(attrNode.value.type.endsWith('Literal')
                ? JSON.stringify(attrNode.value.value)
                : code.substring(attrNode.value.expression.start + offset, attrNode.value.expression.end + offset), "}");
        }
    })
        .join(','), ")} ").concat(node.selfClosing ? '/' : '', ">");
    code = code.substring(0, start) + transpiledStatement + code.substring(end);
    offset = transpiledStatement.length - statement.length;
    return { code: code, offset: offset };
}

var buildStyles = function (styles) {
    var _styles = '';
    if (styles) {
        styles.forEach(function (it) {
            if (it) {
                _styles += it;
            }
        });
    }
    if (_styles !== '') {
        return _styles.trim();
    }
    return undefined;
};
function getSingleFileComponentParts(code) {
    var parts = parseComponent(code);
    if (parts.script) {
        parts.script = parts.script.replace(/\/\*[\s\S]*?\*\/|([^:]|^)\/\/.*$/gm, '$1');
    }
    return parts;
}
function injectTemplateAndParseExport(parts, config) {
    var templateString = parts.template
        ? parts.template.replace(/`/g, '\\`').replace(/\$\{/g, '\\${')
        : undefined;
    if (!parts.script) {
        return { component: "{template: `".concat(templateString, "` }") };
    }
    var comp = parseScriptCode(parts.script, config);
    if (templateString) {
        comp.component = "{template: `".concat(templateString, "`, ").concat(comp.component, "}");
    }
    else {
        comp.component = "{".concat(comp.component, "}");
    }
    return comp;
}
function parseScriptCode(code, config) {
    if (config === void 0) { config = {}; }
    var preprocessing = '';
    var startIndex = -1;
    var endIndex = -1;
    var offset = 0;
    var renderFunctionStart = -1;
    var ast = getAst(code).program;
    var isFunctional = false;
    var setFunctionalComponent = function (node) {
        if (['ArrowFunctionExpression', 'FunctionDeclaration'].includes(node.type)) {
            isFunctional = true;
        }
    };
    walkes(ast, {
        // export const MyComponent = {}
        ExportNamedDeclaration: function (node) {
            preprocessing = code.slice(0, node.start + offset);
            startIndex = node.declaration.declarations[0].init.start + offset;
            endIndex = node.declaration.declarations[0].init.end + offset;
            if (node.declarations) {
                renderFunctionStart = getRenderFunctionStart(node.declarations[0]);
            }
            setFunctionalComponent(node.declaration.declarations[0]);
        },
        //export default {}
        ExportDefaultDeclaration: function (node) {
            preprocessing = code.slice(0, node.start + offset);
            startIndex = node.declaration.start + offset;
            endIndex = node.declaration.end + offset;
            renderFunctionStart = getRenderFunctionStart(node.declaration);
            setFunctionalComponent(node.declaration);
        },
        //module.exports = {}
        AssignmentExpression: function (node) {
            if (/exports/.test(node.left.name) ||
                (node.left.object &&
                    /module/.test(node.left.object.name) &&
                    /exports/.test(node.left.property.name))) {
                preprocessing = code.slice(0, node.start + offset);
                startIndex = node.right.start + offset;
                endIndex = node.right.end + offset;
                setFunctionalComponent(node.right);
            }
        }
    });
    walkes(ast, {
        JSXOpeningElement: function (node) {
            if (node.attributes.some(function (attrNode) { return attrNode.type === 'JSXSpreadAttribute'; })) {
                var ret = transformOneJSXSpread(node, code, offset, config);
                if (node.start + offset < startIndex) {
                    offset += ret.offset;
                }
                else if (node.end + offset < endIndex) {
                    endIndex += ret.offset;
                }
                code = ret.code;
            }
        }
    });
    if (startIndex === -1) {
        throw new Error('Failed to parse single file component: ' + code);
    }
    if (renderFunctionStart > 0 && !isVue3) {
        renderFunctionStart += offset;
        code = insertCreateElementFunction(code.slice(0, renderFunctionStart + 1), code.slice(renderFunctionStart + 1));
        endIndex += JSX_ADDON_LENGTH;
    }
    var component = isFunctional
        ? "render: ".concat(code.slice(startIndex, endIndex))
        : code.slice(startIndex + 1, endIndex - 1);
    return {
        preprocessing: preprocessing,
        component: component,
        postprocessing: code.slice(endIndex)
    };
}
var JSX_ADDON_LENGTH = 31;
function getRenderFunctionStart(objectExpression) {
    if (objectExpression && objectExpression.properties) {
        var nodeProperties = objectExpression.properties;
        var renderFunctionObj = nodeProperties.find(function (p) { return p.key && p.key.type === 'Identifier' && p.key.name === 'render'; });
        if (renderFunctionObj === null || renderFunctionObj === void 0 ? void 0 : renderFunctionObj.body) {
            return renderFunctionObj.body.start;
        }
    }
    return -1;
}
function insertCreateElementFunction(before, after) {
    return "".concat(before, ";const h = this.$createElement;").concat(after);
}
/**
 * Coming out of this function all SFC should be in the `new Vue()` format
 * it should as well have been stripped of exports and all imports should have been
 * transformed into requires
 */
function normalizeSfcComponent(code, config) {
    if (config === void 0) { config = {}; }
    var parts = getSingleFileComponentParts(code);
    var _a = injectTemplateAndParseExport(parts, config), preprocessing = _a.preprocessing, component = _a.component;
    return {
        script: [preprocessing, "return ".concat(component)].join(';'),
        style: buildStyles(parts.styles)
    };
}

/**
 * Reads the code in string and separates the javascript part and the html part
 * then sets the nameVarComponent variable with the value of the component parameters
 * @param code
 * @param config sucrase config to be used when transforming
 *
 */
function compileVueCodeForEvalFunction(code, config) {
    if (config === void 0) { config = {}; }
    var nonCompiledComponent = prepareVueCodeForEvalFunction(code, config);
    var configWithTransforms = {
        production: true,
        jsxPragma: config.jsxPragma,
        jsxFragmentPragma: config.jsxFragmentPragma,
        transforms: __spreadArray(['typescript', 'imports'], (config.jsxPragma ? ['jsx'] : []), true)
    };
    return __assign(__assign({}, nonCompiledComponent), { script: transform(nonCompiledComponent.script, configWithTransforms).code });
}
function prepareVueCodeForEvalFunction(code, config) {
    var style;
    var vsgMode = false;
    var template;
    // if the component is written as a Vue sfc,
    // transform it in to a "return"
    // even if jsx is used in an sfc we still use this use case
    if (isCodeVueSfc(code)) {
        return normalizeSfcComponent(code, config);
    }
    // if it's not a new Vue, it must be a simple template or a vsg format
    // lets separate the template from the script
    if (!/new Vue\(/.test(code)) {
        // this for jsx examples without the SFC shell
        // export default {render: (h) => <Button>}
        if (config.jsxPragma) {
            var _a = parseScriptCode(code, config), preprocessing = _a.preprocessing, component = _a.component;
            return {
                script: "".concat(preprocessing, ";return {").concat(component, "};")
            };
        }
        var findStartTemplateMatch = /^\W*</.test(code) ? { index: 0 } : code.match(/\n[\t ]*</);
        var limitScript = findStartTemplateMatch && findStartTemplateMatch.index !== undefined
            ? findStartTemplateMatch.index
            : -1;
        template = limitScript > -1 ? code.slice(limitScript) : undefined;
        code = limitScript > -1 ? code.slice(0, limitScript) : code;
        vsgMode = true;
    }
    var ast = getAst(code);
    var offset = 0;
    var varNames = [];
    walkes(ast.program, __assign({ 
        // replace `new Vue({data})` by `return {data}`
        ExpressionStatement: function (node) {
            if (node.expression.type === 'NewExpression' && node.expression.callee.name === 'Vue') {
                var before = code.slice(0, node.expression.start + offset);
                var optionsNode = node.expression.arguments && node.expression.arguments.length
                    ? node.expression.arguments[0]
                    : undefined;
                var renderIndex = getRenderFunctionStart(optionsNode);
                var endIndex = optionsNode.end;
                if (renderIndex > 0 && !isVue3) {
                    code = insertCreateElementFunction(code.slice(0, renderIndex + 1), code.slice(renderIndex + 1));
                    endIndex += JSX_ADDON_LENGTH;
                }
                var after = optionsNode ? code.slice(optionsNode.start + offset, endIndex + offset) : '';
                code = before + ';return ' + after;
            }
        }, 
        // transform all imports into require function calls
        ImportDeclaration: function (node) {
            if (vsgMode && node.specifiers) {
                node.specifiers.forEach(function (s) { return varNames.push(s.local.name); });
            }
        } }, (vsgMode
        ? {
            VariableDeclaration: function (node) {
                node.declarations.forEach(function (declaration) {
                    if (declaration.id.name) {
                        // simple variable declaration
                        varNames.push(declaration.id.name);
                    }
                    else if (declaration.id.properties) {
                        // spread variable declaration
                        // const { all:names } = {all: 'foo'}
                        declaration.id.properties.forEach(function (p) {
                            varNames.push(p.value.name);
                        });
                    }
                });
            },
            FunctionDeclaration: function (node) {
                varNames.push(node.id.name);
            }
        }
        : {})));
    if (vsgMode) {
        code += ";return {data:function(){return {".concat(
        // add local vars in data
        // this is done through an object like {varName: varName}
        // since each varName is defined in compiledCode, it can be used to init
        // the data object here
        varNames.map(function (varName) { return "".concat(varName, ":").concat(varName); }).join(','), "};}}");
    }
    return {
        script: code,
        style: style,
        template: template
    };
}

function getImports(code) {
    var imports = [];
    try {
        var ast = getAst(code);
        walkes(ast.program, {
            ImportDeclaration: function (node) {
                imports.push(node.source.value);
            },
            CallExpression: function (node) {
                if (node.callee.name === 'require' && node.arguments[0].value) {
                    imports.push(node.arguments[0].value);
                }
            }
        });
        return imports;
    }
    catch (e) {
        return [];
    }
}

export { compileVueCodeForEvalFunction as compile, getImports };
